*-----------------------------------------------------------------------------*

	MODULE	SPEEDTEST
	TEXT
	OPT	P=68040,P=68881
	OPT	O+

; cache stuff from DML

*-----------------------------------------------------------------------------*
*	68030(+020) cache control
*-----------------------------------------------------------------------------*

; bits (used by 68020/30)
ienab030_bit	=	0
ifrz030_bit	=	1
iclre030_bit	=	2
iclr030_bit	=	3
ibrst030_bit	=	4

denab030_bit	=	8
dfrz030_bit	=	9
dclre030_bit	=	10
dclr030_bit	=	11
dbrst030_bit	=	12
wa030_bit	=	13

; flags (used by 68020/30)
ienab030	=	1<<ienab030_bit
ifrz030		=	1<<ifrz030_bit
iclre030	=	1<<iclre030_bit
iclr030		=	1<<iclr030_bit
ibrst030	=	1<<ibrst030_bit

denab030	=	1<<denab030_bit
dfrz030		=	1<<dfrz030_bit
dclre030	=	1<<dclre030_bit
dclr030		=	1<<dclr030_bit
dbrst030	=	1<<dbrst030_bit
wa030		=	1<<wa030_bit

; masks (valid flags for 68020/30)
i030_cache_mask	=	denab030|dfrz030|dclre030|dclr030|dbrst030|wa030	; all d_cache bit are set and all i cache bit are cleared so we can do a AND
d030_cache_mask	=	ienab030|ifrz030|iclre030|iclr030|ibrst030		; all i_cache bit are set and all d cache bit are cleared so we can do a AND

*-----------------------------------------------------------------------------*
*	68040 cache control
*-----------------------------------------------------------------------------*

ienab040_bit	=	15
denab040_bit	=	31

ienab040	=	1<<ienab040_bit
denab040	=	1<<denab040_bit

cache_mask_040	=	ienab040|denab040

*-----------------------------------------------------------------------------*
*	68060 cache control
*-----------------------------------------------------------------------------*

denab060_bit	=	31
nad060_bit	=	30
esb060_bit	=	29
dpi060_bit	=	28
foc060_bit	=	27

ebc060_bit	=	23
cabc060_bit	=	22
cubc060_bit	=	21

ienab060_bit	=	15
nai060_bit	=	14
fic060_bit	=	13

ienab060	=	1<<ienab060_bit
nad060		=	1<<nad060_bit
esb060		=	1<<esb060_bit
dpi060		=	1<<dpi060_bit
foc060		=	1<<foc060_bit
ebc060		=	1<<ebc060_bit
cabc060		=	1<<cabc060_bit
cubc060		=	1<<cubc060_bit
denab060	=	1<<denab060_bit
nai060		=	1<<nai060_bit
fic060		=	1<<fic060_bit

cache_mask_060	=	ienab060|nad060|esb060|dpi060|foc060|ebc060|cabc060|cubc060|denab060|nai060|fic060

*-----------------------------------------------------------------------------*

timerb		=	$120
timerc		=	$114

tbcr		=	$fffffa1b
cdcr		=	$fffffa1d
tbdr		=	$fffffa21
tcdr		=	$fffffa23
iera		=	$fffffa07
ierb		=	$fffffa09
imra		=	$fffffa13
imrb		=	$fffffa15
isra		=	$fffffa0f
isrb		=	$fffffa11

*-----------------------------------------------------------------------------*

; ***IMPORTANT*** update this when modifying the function table
max_functions	=	27

; use 10K block
blksize		=	10240

; mode for STRAM
stram_mode	= 	0
ttram_mode	= 	1


;teststub

	ifd		teststub

*-------------------------------------------------------*
test:	move.l		4(sp),a5
;	lea		128(a5),a4	; command string
;	move.l		a4,cli
*-------------------------------------------------------*
*	Mshrink						*
*-------------------------------------------------------*
	move.l		12(a5),d0	; text segment
	add.l		20(a5),d0	; data segment
	add.l		28(a5),d0	; bss segment
	add.l		#256,d0		; base page
;	lea		usp_stack,sp
	move.l		d0,-(sp)
	move.l		a5,-(sp)
	clr.w		-(sp)
	move.w		#$4a,-(sp)
	trap		#1
	lea		12(sp),sp

;	bra		*

	lea		test_params,a1
	bsr		code_start
	
	clr.w		-(sp)
	trap		#1
	
	
test_params:
	dc.l		9
	dc.l		1000	

	endc

*-----------------------------------------------------------------------------*
code_start:
*-----------------------------------------------------------------------------*
	bra.s		modstart
*-----------------------------------------------------------------------------*

d_cache_state
	dc.w		0
i_cache_state
	dc.w		0
bas_params
	dc.l		0
processor
	dc.l		68000
starttime
	dc.l		0
sys_state
	ds.l		2

*-----------------------------------------------------------------------------*
modstart:
*-----------------------------------------------------------------------------*
	; save registers
	movem.l		d0-d7/a0-a6,-(sp)
	; get params
	lea		bas_params(pc),a0
	move.l		4(a1),(a0)				; parameter
	; fetch function index
	move.l		(a1),d0					; command
	; safety check
	cmp.w		#max_functions-1,d0
	bhi		exit
	; look up function relative offset
	add.w		d0,d0
	move.w		.func_table(pc,d0.w),d0
	; calculate function absolute address
	lea		.func_table(pc,d0.w),a2
	; execute function in supervisor mode
	pea 		(a2)
	move.w 		#38,-(sp)
	trap 		#14
	addq.l 		#6,sp
	; complete
	bra		exit

.func_table:

	; integer
	dc.w		bitlogic_test-.func_table		; 0	68000-60
	dc.w		branch_test-.func_table			; 1	68000-60
	dc.w		mul_16_32_test-.func_table		; 2	68000-60
	dc.w		div_32_16_test-.func_table		; 3	68000-60
	dc.w		mul_32_64_test-.func_table		; 4	68020-40 (emulated on 060 via M68060ISP driver software)
	dc.w		div_64_32_test-.func_table		; 5	68020-40 (emulated on 060 via M68060ISP driver software)

	; float math
	dc.w		fprim_test-.func_table			; 6	68881/2/040/060
	dc.w		fpow_test-.func_table			; 7	68881/2 (emulated on 040-60 via M680x0FPSP driver software)
	dc.w		ftrig_test-.func_table			; 8	68881/2 (emulated on 040-60 via M680x0FPSP driver software)

	; memory
	dc.w		tos_r_test-.func_table			; 9
	dc.w		ram_rw_test-.func_table			; 10
	dc.w		stram_rw_test-.func_table		; 11
	dc.w		ttram_rw_test-.func_table		; 12
	dc.w		vram_rw_test-.func_table		; 13

	; cache control
	dc.w		get_cache_state-.func_table		; 14	68020-30
	dc.w		d_cache_on-.func_table			; 15	68030 (no d-cache on 020, ignored)
	dc.w		d_cache_off-.func_table			; 16	68030 (no d-cache on 020, ignored)
	dc.w		i_cache_on-.func_table			; 17	68020-30
	dc.w		i_cache_off-.func_table			; 18	68020-30

	; sys timing control
	dc.w		sys_sleep-.func_table			; 19	68000-60
	dc.w		sys_wake-.func_table			; 20	68000-60
	dc.w		sys_tstart-.func_table			; 21	68000-60
	dc.w		sys_tstop-.func_table			; 22	68000-60

	; processor
	dc.w		sys_setcpu-.func_table			; 23	68000-60

	; misc.
	dc.w		sys_acia_off-.func_table		; 24	68000-60
	dc.w		sys_acia_on-.func_table			; 25	68000-60
	dc.w		sys_acia_reset-.func_table		; 26	68000-60

exit:	; restore registers
	movem.l 	(sp)+,d0-d7/a0-a6
	rts




*-----------------------------------------------------------------------------*
*	Bitwise logic
*-----------------------------------------------------------------------------*
bitlogic_test:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
.bit_outer_loop:
	move.l		#%00101101010011100100110101001011,d0
	move.w		#(1024/8)-1,d1 ; 1024 ops with 8 per loop
.bit_loop:
	;
	eor.l		d1,d3
	sub.l		d2,d4
	or.l		d0,d5
	rol.l		#3,d0
	and.l		d1,d3
	add.l		d2,d4
	eor.l		d0,d5
	addx.l		d3,d0
	;
	dbra		d1,.bit_loop
	dbra 		d2,.bit_outer_loop
.err:	rts


*-----------------------------------------------------------------------------*
*	Conditional branching
*-----------------------------------------------------------------------------*
branch_test:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
.bcc_outer_loop:
	move.w		#(1024/6)-1,d1 ; approximate 1024 branches with ~6 per loop
	move.l		#%00101101010011100100110101001011,d0
	add.l		d0,d0
.bcc_loop:
	;
	bcc.s		.ta
	bne.s		.tb
.ta:	addx.l		d0,d0
.tb:
	bhi.s		.tc
	beq.s		.td
.tc:	addx.l		d0,d0
.td:
	bls.s		.te
	bne.s		.tf
.te:	addx.l		d0,d0
.tf:
	bvc.s		.tg
	beq.s		.th
.tg:	addx.l		d0,d0
.th:
	;
	dbra		d1,.bcc_loop
	dbra 		d2,.bcc_outer_loop
.err:	rts


*-----------------------------------------------------------------------------*
*	Integer DIV 32/16=16r:16 speed test
*-----------------------------------------------------------------------------*
div_32_16_test:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	moveq		#3,d7
	move.l		#$0001FFFF,d6
.div_outer_loop:
	move.l		d6,d3
	moveq		#(1024/16)-1,d1 ; 1024 loops
.div_loop:
	;
	rept		16
	divu.w		d7,d3
	endr
	;
	dbra		d1,.div_loop
	dbra 		d2,.div_outer_loop
.err:	rts

*-----------------------------------------------------------------------------*
*	Integer MUL 16*16=32 speed test
*-----------------------------------------------------------------------------*
mul_16_32_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	move.w		#%0100110101001011,d7
.mul_outer_loop:
	move.l		#$0000FFFF,d5
	move.w		#(1024/16)-1,d1 ; 1024 loops
.mul_loop:
	;
	move.l		d5,d3
	rept		16
	mulu.w		d7,d3
	endr
	;
	dbra		d1,.mul_loop
	dbra 		d2,.mul_outer_loop
.err:	rts

*-----------------------------------------------------------------------------*
*	Integer DIV 64/32=32r:32 speed test
*-----------------------------------------------------------------------------*
div_64_32_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	moveq		#3,d7
.div_outer_loop:	
	moveq		#1,d5
	moveq		#-1,d6
	move.w		#(1024/16)-1,d1 ; 1024 loops
.div_loop:
	;
	move.l		d5,d3
	move.l		d6,d4
	rept		16
	divu.l		d7,d3:d4
	endr
	;
	dbra		d1,.div_loop
	dbra 		d2,.div_outer_loop
.err:	rts

*-----------------------------------------------------------------------------*
*	Integer MUL 32*32=64 speed test
*-----------------------------------------------------------------------------*
mul_32_64_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	move.l		#%00101101010011100100110101001011,d7
.mul_outer_loop:
	moveq		#0,d3
	move.l		#$0000FFFF,d5
	move.w		#(1024/16)-1,d1 ; 1024 loops
.mul_loop:
	;
	move.l		d5,d4
	rept		16
	mulu.l		d7,d3:d4
	endr
	;
	dbra		d1,.mul_loop
	dbra 		d2,.mul_outer_loop
.err:	rts




*-----------------------------------------------------------------------------*
*	Float FMUL,FDIV,FADD,FNEG,FSQRT,FCMP,FABS,FREM combined speed test
*-----------------------------------------------------------------------------*
fprim_test:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	fmovem		fp0-fp7,-(sp)
.fprim_outer_loop:
	fmove.s		#1.001,fp0
	fmove.x		fp0,fp1
	fmove.x		fp0,fp2
	fmove.x		fp0,fp3
	fmove.x		fp0,fp4
	fmove.x		fp0,fp5
	move.w		#(1024/8)-1,d1 ; 1024 loops
.fprim_loop:
	;
	fmul.x		fp2,fp3
	fdiv.x		fp4,fp5
	fadd.x		fp0,fp1
	fneg.x		fp5
	fsqrt.x		fp3
	fcmp.x		fp4,fp5
	fabs.x		fp1
	frem.x		fp3,fp4
	;
	dbra		d1,.fprim_loop
	dbra 		d2,.fprim_outer_loop
	fmovem		(sp)+,fp0-fp7
.err:	rts

*-----------------------------------------------------------------------------*
*	Float pow() (FLOG+FETOX) composite speed test
*-----------------------------------------------------------------------------*
fpow_test:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	fmovem		fp0-fp7,-(sp)
.fpow_outer_loop:
	fmove.s		#1.001,fp0
	fmove.x		fp0,fp1
	fmove.x		fp0,fp2
	move.w		#(1024/2)-1,d1 ; 1024 loops
.fpow_loop:
	;
	fmove.x		fp2,fp0
	flogn.x		fp0
        fmul.x		fp1,fp0
        fetox.x		fp0
        ;
	fmove.x		fp2,fp0
	flogn.x		fp0
        fmul.x		fp1,fp0
        fetox.x		fp0
	;
	dbra		d1,.fpow_loop
	dbra 		d2,.fpow_outer_loop
	fmovem		(sp)+,fp0-fp7
.err:	rts

*-----------------------------------------------------------------------------*
*	Float FTRIG (FSIN,FCOS,FTAN) speed test
*-----------------------------------------------------------------------------*
ftrig_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	fmovem		fp0-fp7,-(sp)
.ftrig_outer_loop:
	move.w		#(1024/8)-1,d1 ; 1024 loops
.ftrig_loop:
	;
	fmove.s		#1.001,fp0
	fmove.x		fp0,fp1
	fsin.x		fp0
	fmove.x		fp1,fp2
	fcos.x		fp1
	fmove.x		fp2,fp3
	ftan.x		fp2
	;
	fmove.x		fp3,fp4
	fasin.x		fp3
	fmove.x		fp4,fp5
	facos.x		fp4
	fmove.x		fp5,fp6
	fatan.x		fp5
	;
	fmove.x		fp6,fp7
	ftanh.x		fp6
	fatanh.x	fp7
	;
	dbra		d1,.ftrig_loop
	dbra 		d2,.ftrig_outer_loop
	fmovem		(sp)+,fp0-fp7
.err:	rts




*-----------------------------------------------------------------------------*
*	TOS read speed test
*-----------------------------------------------------------------------------*
tos_r_test:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	subq.l		#1,d2
	bmi.s		.err
	; get TOS base
	move.l		$4f2.w,a0

.tos_outer_loop:
	; reads
	movea.l		a0,a6
	move.w		#(blksize/(8*8*4))-1,d1
.tos_r_loop:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	dbra		d1,.tos_r_loop
	dbra 		d2,.tos_outer_loop
.err:	rts


*-----------------------------------------------------------------------------*
* 	default RAM access speed test (read and write)
*-----------------------------------------------------------------------------*
ram_rw_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	; Malloc
	; return block address in D0
	move.l		#blksize+16,-(sp)
	move.w		#$48,-(sp)
	trap 		#1
	addq.l		#6,sp
	; if there is no memory available, this return 0
	move.l		d0,a5
	tst.w		d0
	bne.s		.cont_ram_test
	rts

.cont_ram_test:
	; align to 16 bytes
	add.l		#16-1,d0
	and.w		#-16,d0
	move.l		d0,a0
.ram_outer_loop:
	;
	; reads
	movea.l		a0,a6
	; blk_size/nb_reg*4 used  = nb loops, here we use 8 registers as 10240/(8*4) is an integer
	; d3, d4, d5, d6, d7, a1, a2, a3
	move.w		#(blksize/(8*8*4))-1,d1
.ram_r_loop:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	dbra		d1,.ram_r_loop

	; writes
;	lea	blksize(a0),a6
	move.w		#(blksize/(8*8*4))-1,d1
.ram_w_loop
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	dbra		d1,.ram_w_loop
	;
	dbra		d2,.ram_outer_loop

	; free
	move.l		a5,-(sp)
	move.w		#$49,-(sp)
	trap		#1
	addq.l		#6,sp
	rts

*-----------------------------------------------------------------------------*
* 	ST-RAM access speed test (read and write)
*-----------------------------------------------------------------------------*
stram_rw_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	; Mxalloc
	; return block address in A0
	move.w		#stram_mode,-(sp)
	move.l		#blksize+16,-(sp)
	move.w		#$44,-(sp)
	trap		#1
	addq.l		#8,sp
	; if there is no memory available, this return 0
	move.l		d0,a5
	tst.w		d0
	bne.s		.cont_ram_test
	rts

.cont_ram_test:
	; align to 16 bytes
	add.l		#16-1,d0
	and.w		#-16,d0
	move.l		d0,a0
.ram_outer_loop:
	;
	; reads
	movea.l		a0,a6
	; blk_size/nb_reg*4 used  = nb loops, here we use 8 registers as 10240/(8*4) is an integer
	; d3, d4, d5, d6, d7, a1, a2, a3
	move.w		#(blksize/(8*8*4))-1,d1
.ram_r_loop:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	dbra		d1,.ram_r_loop

	; writes
;	lea	blksize(a0),a6
	move.w		#(blksize/(8*8*4))-1,d1
.ram_w_loop
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	dbra		d1,.ram_w_loop
	;
	dbra		d2,.ram_outer_loop

	; free
	move.l		a5,-(sp)
	move.w		#$49,-(sp)
	trap		#1
	addq.l		#6,sp
	rts

*-----------------------------------------------------------------------------*
* 	TT-RAM access speed test (read and write)
*-----------------------------------------------------------------------------*
ttram_rw_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	; Mxalloc
	; return block address in A0
	move.w		#ttram_mode,-(sp)
	move.l		#blksize+16,-(sp)
	move.w		#$44,-(sp)
	trap		#1
	addq.l		#8,sp
	; if there is no memory available, this return 0
	move.l		d0,a5
	tst.w		d0
	bne.s		.cont_ram_test
	rts

.cont_ram_test:
	; align to 16 bytes
	add.l		#16-1,d0
	and.w		#-16,d0
	move.l		d0,a0
.ram_outer_loop:
	;
	; reads
	movea.l		a0,a6
	; blk_size/nb_reg*4 used  = nb loops, here we use 8 registers as 10240/(8*4) is an integer
	; d3, d4, d5, d6, d7, a1, a2, a3
	move.w		#(blksize/(8*8*4))-1,d1
.ram_r_loop:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	dbra		d1,.ram_r_loop

	; writes
;	lea		blksize(a0),a6
	move.w		#(blksize/(8*8*4))-1,d1
.ram_w_loop
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	dbra		d1,.ram_w_loop
	;
	dbra		d2,.ram_outer_loop

	; free
	move.l		a5,-(sp)
	move.w		#$49,-(sp)
	trap		#1
	addq.l		#6,sp
	rts



*-----------------------------------------------------------------------------*
* 	Video RAM access speed test (read and write)
*-----------------------------------------------------------------------------*
vram_rw_test
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	; Malloc scratch
	; return block address in D0
	move.l		#blksize,-(sp)
	move.w		#$48,-(sp)
	trap 		#1
	addq.l		#6,sp
	; if there is no memory available, this return 0
	move.l		d0,a5
	tst.w		d0
	bne.s		.cont_ram_test
	rts


.cont_ram_test:
	; Physbase
	move.w		#2,-(sp)
	trap		#14
	addq.l		#2,sp
	move.l		d0,a4

	; backup display
	movea.l		a4,a6
	move.l		a5,a0
	move.l		#(blksize/(8*4))-1,d1
.store:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	d3-d7/a1-a3,(a0)
	lea		8*4(a0),a0
	dbra		d1,.store

.ram_outer_loop:
	;
	; reads
	movea.l		a4,a6
	; blk_size/nb_reg*4 used  = nb loops, here we use 8 registers as 10240/(8*4) is an integer
	; d3, d4, d5, d6, d7, a1, a2, a3
	move.w		#(blksize/(8*8*4))-1,d1
.ram_r_loop:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	(a6)+,d3-d7/a1-a3
	dbra		d1,.ram_r_loop

	; writes
;	lea		blksize(a0),a6
	move.w		#(blksize/(8*8*4))-1,d1
.ram_w_loop
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	movem.l		d3-d7/a1-a3,-(a6)
	dbra		d1,.ram_w_loop
	;
	dbra		d2,.ram_outer_loop

	; restore display
	move.l		a5,a6
	movea.l		a4,a0
	move.l		#(blksize/(8*4))-1,d1
.restore:
	movem.l 	(a6)+,d3-d7/a1-a3
	movem.l 	d3-d7/a1-a3,(a0)
	lea		8*4(a0),a0
	dbra		d1,.restore

	; free scratch
	move.l		a5,-(sp)
	move.w		#$49,-(sp)
	trap		#1
	addq.l		#6,sp
	rts


*-----------------------------------------------------------------------------*
d_cache_on:
*-----------------------------------------------------------------------------*
;	move.l		bas_params(pc),d0
;	lea		processor(pc),a1
;	move.l		d0,(a1)

	lea		processor(pc),a1
	move.l		(a1),d0

	cmp.l		#68000,d0
	beq		get_cache_state_
	
	cmp.l		#68060,d0
	beq.s		.proc060
	cmp.l		#68040,d0
	beq.s		.proc040
.proc02030:
	movec		cacr,d2
	andi.w		#d030_cache_mask,d2		; masks dcache bits
	ori.w		#denab030|dbrst030|dclr030|wa030,d2	; enables dcache
	movec		d2,cacr
	bra		get_cache_state_
.proc060:
	movec		cacr,d2
	or.l		#denab060,d2
	movec		d2,cacr
	bra		get_cache_state_
.proc040:
	movec		cacr,d2
	or.l		#denab040,d2
	movec		d2,cacr
	bra		get_cache_state_


*-----------------------------------------------------------------------------*
d_cache_off:
*-----------------------------------------------------------------------------*
;	move.l		bas_params(pc),d0
;	lea		processor(pc),a1
;	move.l		d0,(a1)

	lea		processor(pc),a1
	move.l		(a1),d0

	cmp.l		#68000,d0
	beq		get_cache_state_

	cmp.l		#68060,d0
	beq.s		.proc060
	cmp.l		#68040,d0
	beq.s		.proc040
.proc02030:
	movec		cacr,d2
	andi.w		#d030_cache_mask,d2		; masks dcache bits
	ori.w		#dclr030,d2  			; disables d_cache
	movec		d2,cacr
	bra		get_cache_state_
.proc060:
	cpusha		dc
	movec		cacr,d2
	and.l		#~denab060,d2
	movec		d2,cacr
	bra		get_cache_state_
.proc040:
	cpusha		dc
	movec		cacr,d2
	and.l		#~denab040,d2
	movec		d2,cacr
	bra		get_cache_state_

*-----------------------------------------------------------------------------*
i_cache_on:
*-----------------------------------------------------------------------------*
;	move.l		bas_params(pc),d0
;	lea		processor(pc),a1
;	move.l		d0,(a1)

	lea		processor(pc),a1
	move.l		(a1),d0

	cmp.l		#68000,d0
	beq		get_cache_state_

	cmp.l		#68060,d0
	beq.s		.proc060
	cmp.l		#68040,d0
	beq.s		.proc040
.proc02030:
	movec		cacr,d2
	andi.w		#i030_cache_mask,d2		; masks icache bits
	ori.w		#ienab030|ibrst030|iclr030,d2	; enables icache
	movec		d2,cacr
	bra		get_cache_state_
.proc060:
	movec		cacr,d2
	or.l		#ienab060,d2
	movec		d2,cacr
	bra		get_cache_state_
.proc040:
	movec		cacr,d2
	or.l		#ienab040,d2
	movec		d2,cacr
	bra		get_cache_state_

*-----------------------------------------------------------------------------*
i_cache_off:
*-----------------------------------------------------------------------------*
;	move.l		bas_params(pc),d0
;	lea		processor(pc),a1
;	move.l		d0,(a1)

	lea		processor(pc),a1
	move.l		(a1),d0

	cmp.l		#68000,d0
	beq		get_cache_state_

	cmp.l		#68060,d0
	beq.s		.proc060
	cmp.l		#68040,d0
	beq.s		.proc040
.proc02030:
	movec		cacr,d2
	andi.w		#i030_cache_mask,d2		; masks icache bits
	ori.w		#iclr030,d2			; enables icache
	movec		d2,cacr
	bra		get_cache_state_
.proc060:
	cpusha		ic
	movec		cacr,d2
	and.l		#~ienab060,d2
	movec		d2,cacr
	bra		get_cache_state_
.proc040:
	cpusha		ic
	movec		cacr,d2
	and.l		#~ienab040,d2
	movec		d2,cacr
	bra		get_cache_state_


*-----------------------------------------------------------------------------*
get_cache_state:
*-----------------------------------------------------------------------------*
;	move.l		bas_params(pc),d0
;	lea		processor(pc),a1
;	move.l		d0,(a1)

	lea		processor(pc),a1
	move.l		(a1),d0
	bsr		get_cache_state_
	
	move.l		bas_params(pc),a0
	lea		d_cache_state(pc),a1
	move.w		(a1)+,(a0)+
	move.w		(a1)+,(a0)+
	rts
	
*-----------------------------------------------------------------------------*
get_cache_state_:
*-----------------------------------------------------------------------------*
	cmp.l		#68000,d0
	beq.s		.err
	
	cmp.l		#68060,d0
	beq.s		.proc060
	cmp.l		#68040,d0
	beq.s		.proc040
.proc02030:
	movec		cacr,d2
	lea		d_cache_state(pc),a3
	move.w		d2,d3
	; mask i_cache bits
	andi.w		#i030_cache_mask,d3
	; copy d_cache_state
	move.w		d3,(a3)
	move.w 		d2,d3
	lea		i_cache_state(pc),a3
	; mask d_cache bits
	andi.w 		#d030_cache_mask,d3
	; copy i_cache_state
	move.w		d3,(a3)
.err:	rts
.proc060:
	movec		cacr,d2
	and.w		#ienab060,d2		; bit #15
	lea		i_cache_state(pc),a3
	move.w		d2,(a3)
	and.l		#denab060,d2		; bit #31 (move to bit #15)
	swap		d2
	lea		d_cache_state(pc),a3
	move.w		d2,(a3)		
	rts
.proc040:
	movec		cacr,d2
	and.w		#ienab040,d2		; bit #15
	lea		i_cache_state(pc),a3
	move.w		d2,(a3)
	and.l		#denab040,d2		; bit #31 (move to bit #15)
	swap		d2
	lea		d_cache_state(pc),a3
	move.w		d2,(a3)		
	rts


*-----------------------------------------------------------------------------*
sys_setcpu:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d0
	lea		processor(pc),a1
	move.l		d0,(a1)
	rts
	

sys_sleep:
	lea		sys_state(pc),a0
	ori.w		#$0700,sr
	move.l		$70.w,(a0)+
	move.l		$114.w,(a0)+
	lea		new_vbi(pc),a1
	move.l		a1,$70.w
	lea		new_tc(pc),a1
	move.l		a1,timerc.w
	andi.w		#$f3ff,sr
	rts

*-----------------------------------------------------------------------------*
sys_wake:
*-----------------------------------------------------------------------------*
	lea		sys_state(pc),a0
	ori.w		#$0700,sr
	move.l		(a0)+,$70.w
	move.l		(a0)+,timerc.w
	andi.w		#$f3ff,sr
	rts

*-----------------------------------------------------------------------------*
sys_tstart:
*-----------------------------------------------------------------------------*
	lea		starttime(pc),a0
	move.l		$4ba.w,d0
.wait:	move.l		$4ba.w,d1
	cmp.l		d1,d0
	beq.s		.wait
	move.l		d1,(a0)
	rts

*-----------------------------------------------------------------------------*
sys_tstop:
*-----------------------------------------------------------------------------*
	move.l		bas_params(pc),d2
	beq.s		.err
	move.l		d2,a1
	move.l		$4ba.w,d0
	sub.l		starttime(pc),d0
	move.l		d0,(a1)	
.err:	rts

*-----------------------------------------------------------------------------*



*-------------------------------------------------------*
_ST_FlushInput:
*-------------------------------------------------------*
	lea		$fffffc00.w,a0
	bra.s		.read
.next:	btst		#0,(a0)
	bne.s		.read
	subq.w		#1,d0
	bne.s		.next
	bra.s		.end
.read:	move.b		2(a0),d0
	move.w		#$12c,d0
	bra.s		.next
.end:	rts

*-------------------------------------------------------*
write_acia:	
*-------------------------------------------------------*
.wait:	btst		#1,$fffffc00.w
	beq.s		.wait
	move.b		d0,$fffffc02.w
	rts

*-------------------------------------------------------*
sys_acia_on:
;_ST_EnableInput:
*-------------------------------------------------------*
	movem.l		d0-a6,-(sp)
	bsr		_ST_FlushInput
	lea		enable_ikbd_seq(pc),a0
.next:	move.b		(a0)+,d0
	bmi.s		.end
	bsr		write_acia
	bra.s		.next
.end:	bsr		_ST_FlushInput
	movem.l		(sp)+,d0-a6
	rts

*-------------------------------------------------------*
sys_acia_off:
;_ST_DisableInput:
*-------------------------------------------------------*
	movem.l		d0-a6,-(sp)
	bsr		_ST_FlushInput
	lea		disable_ikbd_seq(pc),a0
.next:	move.b		(a0)+,d0
	bmi.s		.end
	bsr		write_acia
	bra.s		.next
.end:	bsr		_ST_FlushInput
	movem.l		(sp)+,d0-a6
	rts

*-------------------------------------------------------*
sys_acia_reset:
;_ST_ResetInput:
*-------------------------------------------------------*
	movem.l		d0-a6,-(sp)
	bsr		_ST_FlushInput
	lea		reset_ikbd_seq(pc),a0
.next:	move.b		(a0)+,d0
	bmi.s		.end
	bsr		write_acia
	bra.s		.next
.end:	bsr		_ST_FlushInput
	movem.l		(sp)+,d0-a6
	rts

*-------------------------------------------------------*

disable_ikbd_seq:	dc.b	$13,$1a,$12,-1			; reporting off, joystick events off, mouse events off 
enable_ikbd_seq:	dc.b	$11,$14,$8,-1			; resume reporting, joystick events on, mouse events on
reset_ikbd_seq:		dc.b	$80,$1,-1
			even

	
*-----------------------------------------------------------------------------*
new_vbi:
*-----------------------------------------------------------------------------*
	addq.l		#1,$462.w
	addq.l		#1,$466.w
	rte

*-----------------------------------------------------------------------------*
new_tc:
*-----------------------------------------------------------------------------*
	addq.l		#1,$4ba.w
	bclr		#5,isrb.w
	rte
